@@ -1,19 +1,21 @@
 {-# LANGUAGE BlockArguments #-}
 {-# LANGUAGE DeriveAnyClass #-}
 {-# LANGUAGE DeriveGeneric #-}
-{-# LANGUAGE DerivingStrategies #-}
+{-# LANGUAGE DerivingVia #-}
 {-# LANGUAGE NamedFieldPuns #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE TypeOperators #-}
+{-# LANGUAGE LambdaCase #-}
 
 -- | Tasty-based test suite for duckdb-simple.
 module Main (main) where
 
 import Control.Applicative ((<|>))
 import Control.Exception (ErrorCall, Exception, SomeException, displayException, fromException, try)
-import Control.Monad (replicateM_)
+import Control.Monad (forM_, replicateM_)
 import qualified Data.ByteString as BS
+import Data.Array (Array, elems, listArray)
 import Data.IORef (atomicModifyIORef', newIORef)
 import Data.Int (Int16, Int32, Int64, Int8)
 import Data.List (sortOn)
@@ -34,19 +36,31 @@ import Data.Time.LocalTime (
 import Data.Word (Word16, Word32, Word64, Word8)
 import Data.Ratio ((%))
 import Database.DuckDB.Simple
-import Database.DuckDB.Simple.FromField (
-    BigNum (..),
-    BitString (..),
-    bsFromBool,
-    DecimalValue (..),
-    Field (..),
-    FieldValue (..),
-    IntervalValue (..),
-    TimeWithZone (..),
-    fromBigNumBytes,
-    returnError,
-    toBigNumBytes,
- )
+import Database.DuckDB.Simple.FromField
+    ( BigNum (..)
+    , BitString (..)
+    , bsFromBool
+    , DecimalValue (..)
+    , Field (..)
+    , FieldValue (..)
+    , IntervalValue (..)
+    , TimeWithZone (..)
+    , fromBigNumBytes
+    , returnError
+    , toBigNumBytes
+    )
+import Database.DuckDB.Simple.LogicalRep
+    ( StructField (..)
+    , StructValue (..)
+    , UnionMemberType (..)
+    , UnionValue (..)
+    )
+import Database.DuckDB.Simple.Generic
+    ( genericFromFieldValue
+    , genericToFieldValue
+    , genericToStructValue
+    , ViaDuckDB (..)
+    )
 import Database.DuckDB.Simple.Ok (Ok (..))
 import GHC.Generics (Generic)
 import Numeric.Natural (Natural)
@@ -67,6 +81,60 @@ data Person = Person
 data WithRemaining = WithRemaining Int Int
     deriving (Eq, Show)
 
+data GenericRecord = GenericRecord
+    { grId :: Int
+    , grName :: Text.Text
+    }
+    deriving stock (Eq, Show, Generic)
+
+data GenericSum
+    = SumInt Int
+    | SumBool Bool
+    | SumPair Int Text.Text
+    deriving stock (Eq, Show, Generic)
+
+data GenericEnum
+    = EnumRed
+    | EnumBlue
+    deriving stock (Eq, Show, Generic)
+
+data ViaUser = ViaUser
+    { vuId :: Int64
+    , vuName :: Text.Text
+    }
+    deriving stock (Eq, Show, Generic)
+    deriving (DuckDBColumnType, ToField, FromField) via (ViaDuckDB ViaUser)
+
+data ViaShape
+    = ViaCircle Double
+    | ViaSquare Double
+    deriving stock (Eq, Show, Generic)
+    deriving (DuckDBColumnType, ToField, FromField) via (ViaDuckDB ViaShape)
+
+-- Edge case test types
+data TwoFieldStruct = TwoFieldStruct Int Text.Text
+    deriving stock (Eq, Show, Generic)
+
+data LargeSumType
+    = LargeC1 Int
+    | LargeC2 Int
+    | LargeC3 Int
+    | LargeC4 Int
+    | LargeC5 Int
+    deriving stock (Eq, Show, Generic)
+    deriving (DuckDBColumnType, ToField, FromField) via (ViaDuckDB LargeSumType)
+
+data EmptyStruct = EmptyStruct
+    deriving stock (Eq, Show, Generic)
+    deriving (DuckDBColumnType, ToField, FromField) via (ViaDuckDB EmptyStruct)
+
+data OptionalFields = OptionalFields
+    { ofInt :: Maybe Int
+    , ofText :: Maybe Text.Text
+    }
+    deriving stock (Eq, Show, Generic)
+    deriving (DuckDBColumnType, ToField, FromField) via (ViaDuckDB OptionalFields)
+
 instance FromRow WithRemaining where
     fromRow = do
         firstVal <- field
@@ -375,7 +443,7 @@ duckdbTypeCastTests =
                         result <- try (query_ conn sql) :: IO (Either SomeException [Only FieldValue])
                         runExpectation castLabel castExpectation result
                         maybe (pure ()) assertFailure castExpectFailureReason
-         in maybe base (\reason -> expectFailBecause reason base) castExpectFailureReason
+         in maybe base (`expectFailBecause` base) castExpectFailureReason
 
 runExpectation :: String -> DuckDBExpectation -> Either SomeException [Only FieldValue] -> Assertion
 runExpectation label expectation outcome =
@@ -436,12 +504,11 @@ duckdbCastCases =
     , successCase "BIGNUM" (quoted bigNumLiteralText) "BIGNUM" (ExpectEquals (FieldBigNum (BigNum bigNumLiteral)))
     , successCase "UUID" (quoted $ UUID.toText uuid) "UUID" (ExpectEquals (FieldUUID uuid))
     , successCase "BIT" (quoted $ Text.pack $ show bits) "BIT" (ExpectEquals (FieldBit bits))
+    , successCase "ARRAY" (quoted "[1,2,3]") "INTEGER[3]" (ExpectEquals (FieldArray arrayElements))
      -- This one is broken upstream, instead of a DuckDBTypeTimeNs, we get a DuckDBType 0
-     , failCaseWith "TIME_NS" (quoted "03:04:05.123456789") "TIME_NS" "TIME_NS decoding unsupported" (expectErrorCallContaining "unsupported DuckDB type")
-      -- not implemented yet
-    , failCaseWith "STRUCT" (quoted "{\"a\":1,\"b\":2}") "STRUCT(a INTEGER, b INTEGER)" "STRUCT decoding unsupported" (expectErrorCallContaining "STRUCT columns are not supported")
-    , failCaseWith "ARRAY" (quoted "[1,2,3]") "INTEGER[3]" "ARRAY decoding unsupported" (expectErrorCallContaining "unsupported DuckDB type")
-    , failCaseWith "UNION" (quoted "1") "UNION(\"value\" INTEGER)" "UNION casts unsupported" (expectSQLErrorContaining "UNION")
+    , failCaseWith "TIME_NS" (quoted "03:04:05.123456789") "TIME_NS" "TIME_NS decoding unsupported" (expectErrorCallContaining "INVALID")
+    , successDirect "STRUCT" "{'a': 1, 'b': 2}" (expectStruct structFields)
+    , successDirect "UNION" "CAST(union_value(a := 42) AS UNION(a INTEGER, b VARCHAR))" (expectUnion 0 (Text.pack "a") (FieldInt32 42))
     -- These can never surface from a query, only internally.
     , failCaseOK "ANY" (quoted "1") "ANY" (expectSQLErrorContaining "ANY")
     , failCaseOK "STRING_LITERAL" (quoted literalText) "STRING_LITERAL" (expectSQLErrorContaining "STRING_LITERAL")
@@ -484,15 +551,45 @@ duckdbCastCases =
             , castExpectFailureReason = Nothing
             }
     expectMapEntries expectedPairs =
-        ExpectSatisfies $ \fieldValue ->
-            case fieldValue of
-                FieldMap actualPairs ->
-                    let normalize pairs = sortOn (mapKey . fst) pairs
-                        mapKey (FieldText txt) = txt
-                        mapKey other = Text.pack (show other)
-                     in assertEqual "map entries" (normalize expectedPairs) (normalize actualPairs)
+        ExpectSatisfies $
+           \case
+             FieldMap actualPairs ->
+                 let normalize = sortOn (mapKey . fst)
+                     mapKey (FieldText txt) = txt
+                     mapKey other = Text.pack (show other)
+                  in assertEqual "map entries" (normalize expectedPairs) (normalize actualPairs)
+             other ->
+                 assertFailure ("expected FieldMap, but saw " <> show other)
+    structFields =
+        [ StructField{structFieldName = Text.pack "a", structFieldValue = FieldInt32 1}
+        , StructField{structFieldName = Text.pack "b", structFieldValue = FieldInt32 2}
+        ]
+    expectStruct expectedFields =
+        ExpectSatisfies $
+            \case
+                FieldStruct StructValue{structValueFields, structValueTypes, structValueIndex} -> do
+                    let actualFields = elems structValueFields
+                        actualTypes = elems structValueTypes
+                        expectedIndex = Map.fromList (zip (map structFieldName expectedFields) [0 ..])
+                        typeNames = map structFieldName actualTypes
+                    assertEqual "struct field names" (map structFieldName expectedFields) (map structFieldName actualFields)
+                    assertEqual "struct field values" (map structFieldValue expectedFields) (map structFieldValue actualFields)
+                    assertEqual "struct type field names" (map structFieldName expectedFields) typeNames
+                    assertEqual "struct index map" expectedIndex structValueIndex
                 other ->
-                    assertFailure ("expected FieldMap, but saw " <> show other)
+                    assertFailure ("expected FieldStruct, but saw " <> show other)
+    expectUnion expectedIndex expectedLabel expectedPayload =
+        ExpectSatisfies $
+            \case
+                FieldUnion UnionValue{unionValueIndex, unionValueLabel, unionValuePayload, unionValueMembers} -> do
+                    assertEqual "union tag index" expectedIndex unionValueIndex
+                    assertEqual "union tag label" expectedLabel unionValueLabel
+                    assertEqual "union payload" expectedPayload unionValuePayload
+                    let actualMembers = elems unionValueMembers
+                        expectedMembers = [Text.pack "a", Text.pack "b"]
+                    assertEqual "union member names" expectedMembers (map unionMemberName actualMembers)
+                other ->
+                    assertFailure ("expected FieldUnion, but saw " <> show other)
     expectSQLErrorContaining :: Text.Text -> DuckDBExpectation
     expectSQLErrorContaining needle =
         ExpectException $ \err ->
@@ -558,6 +655,8 @@ duckdbCastCases =
         DecimalValue{decimalWidth = 18, decimalScale = 4, decimalInteger = 123456789}
     listElements =
         [FieldInt32 1, FieldInt32 2, FieldInt32 3]
+    arrayElements =
+        listArray (0, 2) listElements
     mapPairs =
         [ (FieldText (Text.pack "a"), FieldInt32 1)
         , (FieldText (Text.pack "b"), FieldInt32 2)
@@ -621,6 +720,18 @@ typeTests =
                 assertEqual "Word64 round-trip" w64 r64
                 [Only (asWord :: Word)] <- query_ conn "SELECT u64 FROM unsigneds"
                 assertEqual "Word from UBIGINT" (fromIntegral w64) asWord
+        , testCase "round-trips fixed-length arrays" $
+            withConnection ":memory:" \conn -> do
+                _ <- execute_ conn "CREATE TABLE arrays (vals INTEGER[3])"
+                let inputs =
+                        [ listArray (0, 2) [1, 2, 3]
+                        , listArray (0, 2) [4, 5, 6]
+                        ]
+                _ <- executeMany conn "INSERT INTO arrays VALUES (?)" (Only <$> inputs)
+                rows <- query_ conn "SELECT vals FROM arrays ORDER BY rowid" :: IO [Only (Array Int Int)]
+                let expected =
+                        fmap Only inputs
+                assertEqual "array round-trip" expected rows
         , testCase "decodes huge integers as Integer" $
             withConnection ":memory:" \conn -> do
                 let hugeValue = 170141183460469231731687303715884105727 :: Integer
@@ -783,6 +894,146 @@ typeTests =
                 _ <- executeMany conn "INSERT INTO bits VALUES (?)" (fmap Only bitValues)
                 rows <- query_ conn "SELECT bit FROM bits ORDER BY rowid" :: IO [Only BitString]
                 assertEqual "bit round-trip" (fmap Only bitValues) rows
+        , testCase "round-trips struct and union parameters" $
+            withConnection ":memory:" \conn -> do
+                _ <- execute_ conn "CREATE TABLE struct_union (s STRUCT(a INT, b INT), u UNION(a INT, b VARCHAR))"
+                [(structVal, unionVal)] <-
+                    query_ conn "SELECT {'a': 1, 'b': 2}, CAST(union_value(a := 42) AS UNION(a INTEGER, b VARCHAR))" :: IO [(StructValue FieldValue, UnionValue FieldValue)]
+                _ <- execute conn "INSERT INTO struct_union VALUES (?, ?)" (structVal, unionVal)
+                rows <- query_ conn "SELECT s, u FROM struct_union" :: IO [(StructValue FieldValue, UnionValue FieldValue)]
+                assertEqual "struct/union round-trip" [(structVal, unionVal)] rows
+        , testCase "generic struct into database" $
+            withConnection ":memory:" \conn -> do
+                _ <- execute_ conn "CREATE TABLE generic_struct (payload STRUCT(grId INT, grName TEXT))"
+                let records = [GenericRecord 1 (Text.pack "alpha"), GenericRecord 2 (Text.pack "beta")]
+                    structParams = traverse genericToStructValue records
+                case structParams of
+                    Nothing -> assertFailure "expected struct encoding"
+                    Just structs -> do
+                        _ <- executeMany conn "INSERT INTO generic_struct VALUES (?)" (map Only structs)
+                        rows <- query_ conn "SELECT payload FROM generic_struct ORDER BY payload.grId" :: IO [Only (StructValue FieldValue)]
+                        let unwrap (Only s) = s
+                            roundTripped = traverse (genericFromFieldValue . FieldStruct . unwrap) rows
+                        roundTripped @?= Right records
+        , testCase "generic struct encode/decode" $ do
+            let rec = GenericRecord 42 (Text.pack "duck")
+            case genericToStructValue rec of
+                Just structVal -> do
+                    map structFieldName (elems (structValueFields structVal)) @?= [Text.pack "grId", Text.pack "grName"]
+                    map structFieldValue (elems (structValueFields structVal)) @?= [FieldInt64 42, FieldText (Text.pack "duck")]
+                    genericFromFieldValue (FieldStruct structVal) @?= Right rec
+                Nothing -> assertFailure "expected struct encoding"
+        , testCase "generic union encode/decode" $ do
+            let sumVal = SumBool True
+            case genericToFieldValue sumVal of
+                FieldUnion uv -> do
+                    map unionMemberName (elems (unionValueMembers uv)) @?=
+                        [ Text.pack "SumInt"
+                        , Text.pack "SumBool"
+                        , Text.pack "SumPair"
+                        ]
+                    unionValueLabel uv @?= Text.pack "SumBool"
+                    genericFromFieldValue (FieldUnion uv) @?= Right sumVal
+                other -> assertFailure ("expected FieldUnion, got " <> show other)
+        , testCase "generic union complex constructor" $ do
+            let sumVal = SumPair 7 (Text.pack "duck")
+            case genericToFieldValue sumVal of
+                FieldUnion uv -> do
+                    unionValueLabel uv @?= Text.pack "SumPair"
+                    case unionValuePayload uv of
+                        FieldStruct StructValue{structValueFields} -> do
+                            map structFieldName (elems structValueFields) @?= [Text.pack "field1", Text.pack "field2"]
+                            map structFieldValue (elems structValueFields) @?=
+                                [FieldInt64 7, FieldText (Text.pack "duck")]
+                        other -> assertFailure ("expected struct payload, got " <> show other)
+                    genericFromFieldValue (FieldUnion uv) @?= Right sumVal
+                other -> assertFailure ("expected FieldUnion, got " <> show other)
+        , testCase "deriving via generic struct round-trip" $
+            withConnection ":memory:" \conn -> do
+                _ <- execute_ conn "CREATE TABLE via_users (payload STRUCT(vuId BIGINT, vuName TEXT))"
+                let users = [ViaUser 10 (Text.pack "alice"), ViaUser 11 (Text.pack "bob")]
+                _ <- executeMany conn "INSERT INTO via_users VALUES (?)" (map Only users)
+                rows <- query_ conn "SELECT payload FROM via_users ORDER BY payload.vuId" :: IO [Only ViaUser]
+                rows @?= map Only users
+        , testCase "deriving via generic union round-trip" $
+            withConnection ":memory:" \conn -> do
+                _ <- execute_ conn "CREATE TABLE via_shapes (shape UNION(ViaCircle STRUCT(field1 DOUBLE), ViaSquare STRUCT(field1 DOUBLE)))"
+                let shapes = [ViaCircle 1.5, ViaSquare 3.0]
+                _ <- executeMany conn "INSERT INTO via_shapes VALUES (?)" (map Only shapes)
+                rows <- query_ conn "SELECT shape FROM via_shapes ORDER BY rowid" :: IO [Only ViaShape]
+                rows @?= map Only shapes
+        , testCase "generic enum null payload" $ do
+            case genericToFieldValue EnumRed of
+                FieldUnion uv -> do
+                    unionValuePayload uv @?= FieldNull
+                    genericFromFieldValue (FieldUnion uv) @?= Right EnumRed
+                other -> assertFailure ("expected FieldUnion, got " <> show other)
+        -- Edge case tests
+        , testCase "struct with multiple basic fields" $ do
+            -- Test a struct with multiple fields
+            let twoField = TwoFieldStruct 42 (Text.pack "text")
+            case genericToFieldValue twoField of
+                FieldStruct sv -> do
+                    genericFromFieldValue (FieldStruct sv) @?= Right twoField
+                other -> assertFailure ("expected FieldStruct for TwoFieldStruct, got " <> show other)
+        , testCase "large sum type with many constructors" $ do
+            let constructors = [LargeC1 1, LargeC2 2, LargeC3 3, LargeC4 4, LargeC5 5]
+            forM_ constructors $ \val -> do
+                case genericToFieldValue val of
+                    FieldUnion uv -> do
+                        genericFromFieldValue (FieldUnion uv) @?= Right val
+                    other -> assertFailure ("expected FieldUnion, got " <> show other)
+        , testCase "empty struct encodes as null" $ do
+            case genericToFieldValue EmptyStruct of
+                FieldNull -> pure ()
+                other -> assertFailure ("expected FieldNull for empty struct, got " <> show other)
+        , testCase "struct with Maybe fields" $ do
+            let withJust = OptionalFields (Just 42) (Just (Text.pack "text"))
+                withNothing = OptionalFields Nothing Nothing
+            case (genericToFieldValue withJust, genericToFieldValue withNothing) of
+                (FieldStruct sv1, FieldStruct sv2) -> do
+                    genericFromFieldValue (FieldStruct sv1) @?= Right withJust
+                    genericFromFieldValue (FieldStruct sv2) @?= Right withNothing
+                (other1, other2) -> assertFailure ("expected FieldStruct, got " <> show (other1, other2))
+        -- Error case tests
+        , testCase "decode wrong field type fails" $ do
+            let wrongType = FieldInt32 42  -- Trying to decode as GenericRecord
+            case genericFromFieldValue wrongType :: Either String GenericRecord of
+                Left err -> assertBool "error mentions STRUCT" (Text.pack "STRUCT" `Text.isInfixOf` Text.pack err)
+                Right _ -> assertFailure "expected decode error for wrong type"
+        , testCase "decode union as struct fails" $ do
+            case genericToFieldValue (SumInt 42) of
+                FieldUnion uv -> do
+                    case genericFromFieldValue (FieldUnion uv) :: Either String GenericRecord of
+                        Left err -> assertBool "error mentions product type" (Text.pack "product type" `Text.isInfixOf` Text.pack err)
+                        Right _ -> assertFailure "expected decode error"
+                other -> assertFailure ("expected FieldUnion, got " <> show other)
+        , testCase "decode struct as union fails" $ do
+            case genericToFieldValue (GenericRecord 1 (Text.pack "test")) of
+                FieldStruct sv -> do
+                    case genericFromFieldValue (FieldStruct sv) :: Either String GenericSum of
+                        Left err -> assertBool "error mentions sum type" (Text.pack "sum type" `Text.isInfixOf` Text.pack err)
+                        Right _ -> assertFailure "expected decode error"
+                other -> assertFailure ("expected FieldStruct, got " <> show other)
+        , testCase "decode struct with wrong field count fails" $
+            withConnection ":memory:" \conn -> do
+                -- Get a real struct with 3 fields from the database
+                _ <- execute_ conn "CREATE TABLE three_field (s STRUCT(a INT, b TEXT, c INT))"
+                [(Only threeField)] <- query_ conn "SELECT {'a': 1, 'b': 'test', 'c': 99}" :: IO [Only (StructValue FieldValue)]
+                -- Try to decode it as GenericRecord which only has 2 fields
+                case genericFromFieldValue (FieldStruct threeField) :: Either String GenericRecord of
+                    Left err -> assertBool "error mentions extra fields" (Text.pack "extra" `Text.isInfixOf` Text.pack err)
+                    Right _ -> assertFailure "expected decode error for extra fields"
+        , testCase "decode union with wrong tag fails" $
+            withConnection ":memory:" \conn -> do
+                -- Get a real union from the database
+                [(Only unionVal)] <-
+                    query_ conn "SELECT CAST(union_value(x := 42) AS UNION(x INT, y VARCHAR))" :: IO [Only (UnionValue FieldValue)]
+                -- Manually create a modified version with an out-of-range tag
+                let wrongTag = unionVal{unionValueIndex = 99, unionValueLabel = Text.pack "Invalid"}
+                case genericFromFieldValue (FieldUnion wrongTag) :: Either String GenericSum of
+                    Left err -> assertBool "error mentions tag or index" (Text.pack "tag" `Text.isInfixOf` Text.pack err || Text.pack "99" `Text.isInfixOf` Text.pack err)
+                    Right _ -> assertFailure "expected decode error for wrong tag"
         ]
 
 streamingTests :: TestTree
